use std::{
    net::TcpListener,
    sync::{
        atomic::{self, AtomicBool},
        Arc,
    },
    thread::available_parallelism,
};

use crate::{
    handlers::handlers::handle_connection,
    handling::routes::{Route, Uri},
    threading::ThreadPool,
};

#[derive(Clone)]
pub struct MountPoint<'a> {
    pub mountpoint: Uri<'a>,
    pub routes: Vec<Route<'a>>,
}

pub struct Config {
    mountpoints: Option<Vec<MountPoint<'static>>>,
    address: TcpListener,
    threadpool: ThreadPool,
}

impl<'a> Config {
    pub fn mount(mut self, mountpoint: Uri<'static>, routes: Vec<Route<'static>>) -> Config {
        let mut mount_message = format!("  >> \x1b[35m{}\x1b[0m\n", mountpoint);
        for (index, route) in routes.iter().enumerate() {
            let indent_sign;
            match index {
                i if i == routes.len() - 1 => indent_sign = "└─",
                _ => indent_sign = "├─",
            };

            mount_message += &format!(
            "     \x1b[35m{indent_sign}\x1b[0m \x1b[36m(\x1b[0m{}\x1b[36m)\x1b[0m \x1b[32m{}\x1b[0m \x1b[34;4m{}\x1b[24m{}\x1b[0m\n",
                route.name.unwrap_or(""),
                route.method,
                mountpoint,
                route.uri
            )
        }

        println!("{mount_message}");
        let mut temp_mountpoints = None;
        std::mem::swap(&mut self.mountpoints, &mut temp_mountpoints);

        if let Some(mut mountpoints) = temp_mountpoints {
            mountpoints.push(MountPoint { mountpoint, routes });
            self.mountpoints = Some(mountpoints);
        } else {
            self.mountpoints = Some(vec![MountPoint { mountpoint, routes }]);
        }
        self
    }
    pub fn launch(self) {
        let running = Arc::new(AtomicBool::new(true));

        // Clone a reference to `running` to pass to the signal handler closure
        let running_for_handler = running.clone();

        // Set up a signal handler for SIGINT
        ctrlc::set_handler(move || {
            // When the user sends a SIGINT signal (e.g. by pressing CTRL+C),
            // set the `running` flag to false to initiate a graceful shutdown
            println!("SIGNIT received, shutting down gracefully");
            running_for_handler.store(false, atomic::Ordering::SeqCst);
        })
        .expect("Error setting Ctrl-C handler");
        for stream in self.address.incoming() {
            if !running.load(atomic::Ordering::SeqCst) {
                break;
            }
            let stream = stream.unwrap();
            let mountpoints = self.mountpoints.clone().unwrap();
            self.threadpool
                .execute(move || handle_connection(stream, mountpoints));
        }
    }
}
pub fn build(ip: &str) -> Config {
    let listener = TcpListener::bind(ip).unwrap();
    let ip = ip.splitn(2, ":").collect::<Vec<&str>>();
    if ip.len() != 2 {
        panic!("Invalid IP Address");
    }
    let port = ip[1];
    let ip = ip[0];
    let workers = available_parallelism().unwrap().get();
    let threadpool = ThreadPool::new(workers);
    println!(
        "\x1b[34m⚙ Configuration\x1b[0m
  >> \x1b[34mIp\x1b[0m: {ip}
  >> \x1b[34mPort\x1b[0m: {port}
  >> \x1b[34mWorkers\x1b[0m: {workers}
\x1b[35m🛪 Mountpoints\x1b[0m"
    );
    Config {
        mountpoints: None,
        address: listener,
        threadpool,
    }
}
